package main

import (
	"bytes"
	"context"
	"crypto/rand"
	"embed"
	"encoding/json"
	"flag"
	"fmt"
	"io"
	"io/fs"
	"net/http"
	"os"
	"path"
	"regexp"
	"sync/atomic"
	"time"

	"log/slog"

	"go.balki.me/anyhttp"
	"go.balki.me/anyhttp/idle"
	"go.balki.me/collage-maker/collage"
)

var (
	imagesDir      string
	collageDir     string
	localAssets    bool
	collageNameGen *nameGen
	imagesDirFs    fs.FS
	listenAddr     string

	//go:embed web
	webFS embed.FS
)

func main() {
	flag.StringVar(&imagesDir, "images-dir", "images", "Sets the images dir")
	flag.StringVar(&collageDir, "collages-dir", "collages", "Sets the collages dir")
	flag.BoolVar(&localAssets, "local-assets", false, "Serve local assets during development")
	flag.StringVar(&listenAddr, "addr", "127.0.0.1:8767", "Web listen address, see https://pkg.go.dev/go.balki.me/anyhttp#readme-address-syntax")

	flag.Parse()

	collageNameGen = NewNameGen()
	imagesDirFs = os.DirFS(imagesDir)
	imagesURLPath := "images"
	collagesPath := "collages"

	addFileServer := func(path, dir string) {
		httpFileServer := http.FileServer(http.Dir(dir))
		http.Handle("/"+path+"/", http.StripPrefix("/"+path, httpFileServer))
	}

	addFileServer(imagesURLPath, imagesDir)
	addFileServer(collagesPath, collageDir)

	if localAssets {
		httpFileServer := http.FileServer(http.Dir("web"))
		http.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
			w.Header().Add("Cache-Control", "no-cache")
			httpFileServer.ServeHTTP(w, r)
		})
	} else {
		indexModTime := time.Now()
		indexHTML := func() io.ReadSeeker {
			indexHTMLContent, err := webFS.ReadFile("web/index.html")
			if err != nil {
				panic(err)
			}
			devOnlyRegex := regexp.MustCompile("\n[^\n]*<!-- DEVONLY[^\n]*")
			return bytes.NewReader(devOnlyRegex.ReplaceAllLiteral(indexHTMLContent, nil))
		}()
		httpFileServer := func() http.Handler {
			webrootFs, err := fs.Sub(webFS, "web")
			if err != nil {
				panic(err)
			}
			return http.FileServer(http.FS(webrootFs))
		}()
		http.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
			if r.URL.Path == "/" {
				http.ServeContent(w, r, "index.html", indexModTime, indexHTML)
			} else {
				httpFileServer.ServeHTTP(w, r)
			}
		})
	}

	http.HandleFunc("/make-collage", func(w http.ResponseWriter, r *http.Request) {
		collageReq := collage.Request{}
		body, err := io.ReadAll(r.Body)
		if err != nil {
			slog.Error("failed to read request body", "error", err)
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
		if err := json.Unmarshal(body, &collageReq); err != nil {
			slog.Error("failed to unmarshal json request", "error", err)
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
		collageFile, err := MakeCollage(&collageReq)
		if err != nil {
			slog.Error("failed to make collage", "error", err)
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
		if _, err := w.Write([]byte(collageFile)); err != nil {
			slog.Error("Failed to write collageFile", "error", err)
		}
	})
	addrType, server, done, err := anyhttp.Serve(listenAddr, idle.WrapHandler(nil))
	if err != nil {
		slog.Error("anyhttp Serve failed", "error", err)
	}

	if addrType == anyhttp.SystemdFD {
		if err := idle.Wait(30 * time.Minute); err != nil {
			slog.Error("Failed to wait for idler", "error", err)
		}
		ctx, cancel := context.WithTimeout(context.Background(), 1*time.Minute) // Don't want any stuck connections
		defer cancel()
		if err := server.Shutdown(ctx); err != nil {
			slog.Error("http server Shutdown failed", "error", err)
		}
	} else {
		<-done
	}
}

type nameGen struct {
	prefix  string
	counter atomic.Uint64
}

func NewNameGen() *nameGen {
	currentTime := time.Now().Unix()
	randBytes := make([]byte, 8)
	_, err := rand.Read(randBytes)
	if err != nil {
		panic(err)
	}
	alpha := []byte("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
	uniqRunID := ""
	for _, b := range randBytes {
		uniqRunID += string(alpha[int(b)%len(alpha)])
	}
	return &nameGen{
		prefix:  fmt.Sprintf("%d-%s", currentTime, uniqRunID),
		counter: atomic.Uint64{},
	}
}

func (n *nameGen) Next() string {
	return fmt.Sprintf("%s-%d", n.prefix, n.counter.Add(1))
}

func MakeCollage(req *collage.Request) (string, error) {
	collageFile := fmt.Sprintf("collage-%s.jpg", collageNameGen.Next())
	out, err := os.Create(path.Join(collageDir, collageFile))
	if err != nil {
		return "", fmt.Errorf("failed to create collage output file, err: %w", err)
	}
	if err := collage.Make(req, imagesDirFs, out); err != nil {
		return "", fmt.Errorf("failed to make collage, err: %w", err)
	}
	return collageFile, nil
}
